<?php

namespace PiiuHelmes\Controller;

use Exception;
use Silex\Application;
use PiiuHelmes\Exception\NoMatchingUserException;
use PiiuHelmes\Exception\NoSectorsException;
use PiiuHelmes\Model\Sector;
use PiiuHelmes\Model\User;
use PiiuHelmes\Repository\SectorRepository;
use PiiuHelmes\Repository\UserRepository;
use Symfony\Component\Form\Extension\Core\Type\CheckboxType;
use Symfony\Component\Form\Extension\Core\Type\ChoiceType;
use Symfony\Component\Form\Extension\Core\Type\FormType;
use Symfony\Component\Form\Extension\Core\Type\SubmitType;
use Symfony\Component\Form\Extension\Core\Type\TextType;
use Symfony\Component\Form\FormFactory;
use Symfony\Component\Form\FormInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Session\Session;
use Symfony\Component\Validator\Constraints;
use Symfony\Component\Validator\Exception\ValidatorException;

class FormController
{
    /**
     * @var array
     */
    protected $errors = [];

    /**
     * @param Application $app
     * @param Request $request
     * @return RedirectResponse|string
     */
    public function indexAction(Application $app, Request $request)
    {
        try {
            $user = $this->getUserFromSession($app['session'], $app['userRepository'], $request);
            $sectors = $this->getSectors($app['sectorRepository']);
        } catch (Exception $e) {
            return $app['twig']->render('error.twig', [
                'error' => 'Error connecting to the database'
            ]);
        }

        $form = $this->getForm($app['form.factory'], $user, $sectors);

        $form->handleRequest($request);
        if ($form->isSubmitted()) {
            try {
                $user = $this->handleSubmittedForm($app['userRepository'], $form, $user);
                $app['session']->set('user_id', $user->getId());
                return $app->redirect(
                    $app['url_generator']->generate('index', [
                        'success' => true
                    ])
                );
            } catch (Exception $e) {
                $this->errors[] = 'Could not save user. Please try again.';
            }
        }

        return $app['twig']->render('index.twig', [
            'isEdit' => $user->hasId(),
            'form' => $form->createView(),
            'errors' => $this->errors,
            'success' => $request->get('success') ?? false
        ]);
    }

    /**
     * @param Session $session
     * @param UserRepository $userRepository
     * @param Request $request
     * @return User
     */
    private function getUserFromSession(Session $session, UserRepository $userRepository, Request $request) : User
    {

        $currentUserId = $session->get('user_id');
        if (!empty($currentUserId)) {
            try {
                return $userRepository->getUserById($currentUserId);
            } catch (NoMatchingUserException $noMatchingUserException) {
                // We're displaying an error, but creating a new user anyway
                if ($request->isMethod('GET')) {
                    $this->errors[] = 'Could not find existing user. Create a new one.';
                }
            }
        }

        return new User();
    }

    /**
     * @param SectorRepository $sectorRepository
     * @return array
     */
    private function getSectors(SectorRepository $sectorRepository) : array
    {
       try {
           $sectors = $sectorRepository->getList();
           return array_reduce($sectors, function ($result, Sector $item) {
               // Making this usable for Symfony forms
               $indent = str_repeat("\t", $item->getIndent());
               $result[$indent.$item->getName()] = $item->getId();
               return $result;
           }, []);
       } catch (NoSectorsException $noSectorsException) {
           $this->errors[] = $noSectorsException->getMessage();
           return[];
       }

    }

    /**
     * @param FormFactory $factory
     * @param User $user
     * @param array $sectors
     * @return FormInterface
     */
    private function getForm(FormFactory $factory, User $user, array $sectors) : FormInterface
    {
        $data = [
            'name' => $user->getName(),
            'sectors' => $user->getSectors(),
            'agree' => $user->isAgree()
        ];

        return $factory->createBuilder(FormType::class, $data)
            ->add('name', TextType::class, [
                'label' => 'Name',
                'required' => true,
                'constraints' => [
                    new Constraints\NotBlank(),
                    new Constraints\Length(['min' => 2])
                ]
            ])
            ->add('sectors', ChoiceType::class, [
                'label' => 'Sectors',
                'required' => true,
                'choices' => $sectors,
                'attr' => [
                    'size' => 5
                ],
                'multiple' => true,
                'constraints' => [
                    new Constraints\NotBlank()
                ]
            ])
            ->add('agree', CheckboxType::class, [
                'label' => 'Agree to terms',
                'required' => true,
                'constraints' => [
                    new Constraints\NotBlank(),
                ]
            ])
            ->add('submit', SubmitType::class, [
                'label' => 'Save',
            ])
            ->getForm();
    }

    /**
     * @param UserRepository $userRepository
     * @param FormInterface $form
     * @param User $user
     * @return User
     */
    private function handleSubmittedForm(UserRepository $userRepository, FormInterface $form, User $user) : User
    {
        if ($form->isValid()) {
            $data = $form->getData();
            $user->setName($data['name']);
            $user->setAgree($data['agree']);
            $user->setSectors($data['sectors']);
            return $userRepository->saveUser($user);
        } else {
            throw new ValidatorException();
        }
    }
}
